@@ -1,18 +1,15 @@
 # Import Fusionbase
-from fusionbase.DataStream import DataStream
-
-# Import pandas
+from fusionbase.Fusionbase import Fusionbase
 import pandas as pd
+import os
 
 # Create a new datastream
 # Provide your API Key and the Fusionbase API URI (usually: https://api.fusionbase.com/api/v1)
-data_stream = DataStream(auth={"api_key": "*** SECRET CREDENTIALS ***"},
+fusionbase = Fusionbase(auth={"api_key": os.getenv('FUSIONBASE_API_KEY')},
                       connection={"base_uri": "https://api.fusionbase.com/api/v1"})
 
 unique_label = 'MUNICH_OKTOBERFEST_BEER_CONSUMPTION'
-
-# Get the datastream by it's unique label
-data_stream_meta = data_stream.get_meta_data_by_label(label=unique_label)
+data_stream_editable = fusionbase.get_datastream(label=unique_label)
 
 # Load the data for which you want to create a datastream
 # Currenty, only lists of dictionaries are supported, i.e., orient="records"
@@ -21,7 +18,4 @@ update_beer_consum_dataset = pd.read_csv(
 
 # Update the data.
 # It will automatically deduplicate the data and only add new entries.
-result = data_stream.update(key=data_stream_meta["_key"], data=update_beer_consum_dataset)
-
-# On success, this will return the id and the key of the newly created data stream
-print(result)
\ No newline at end of file
+result = data_stream_editable.update(data=update_beer_consum_dataset)
\ No newline at end of file
